;
; Purpose:
;	To make a simple node that implements S.N.A.P in AVR-assembler
;
; Author:
;	Gran Nordahl, e99_gno@e.kth.se
;
; Credits:
;	High Tech Horizon, www.hth.com [S.N.A.P]
;	John C. Wren [CRC16]
;
; File, Version, Date:
;	http://www.e.kth.se/~e99_gno/snap/snap01.asm, 1.0, 2000-07-24
;
; Target:
;	AT90S8515/STK200, but will probably work on any AVR with an UART
;
; Description:
;	DB1 of the received packet is shown on the LEDs of STK200 (PORTB).
;	ACK is sent with the accumulated number of ACKs in DB1
;
; Accepted packet structure:
;	HDB2: 1 byte destination/source adress, No flags, ACK request
;	HDB1: Command mode not supported, CRC16, 1 data byte
;
; Support:
;	Ask kindly
;
; Terms of use:
;	For non-commercial purposes you may use this code as you want.
;
;
; For more information about S.N.A.P, go to http://www.hth.com/snap/
;



.include "8515def.inc"

.def	ACK_COUNT = r11
.def	SAB1	= r12
.def	DB1	= r13
.def	CRC_H	= r14
.def	CRC_L	= r15
.def	TEMP	= r16
.def	COUNT	= r17
.def	DATA	= r18

.set	ME	= 2		;Nodes adress
.set	SYNC	= 0b01010100	;Syncronisation byte
.set	HDB2	= 0b01010001	;1 byte destination/source adress, No flags, ACK request.
.set	HDB1	= 0b01000001	;Command mode not supported, CRC16, 1 data byte.
.set	ACK_HDB2= 0b01010010	;1 byte destination/source adress, No flags, ACK response.
.set	ACK_HDB1= 0b01000001	;Command mode not supported, CRC16, 1 data byte.
.set	NAK_HDB2= 0b01010011	;1 byte destination/source adress, No flags, NAK response.
.set	NAK_HDB1= 0b01000000	;Command mode not supported, CRC16, 0 data bytes.

	

;Init STACK
	ldi	TEMP, low(RAMEND)
	out	SPL, TEMP
	ldi	TEMP, high(RAMEND)
	out	SPH, TEMP

;Init UART
	ldi	TEMP, 103	;103 -> 2400bps, 51 -> 4800bps,
	out	UBRR, TEMP	; 25 -> 9600bps, 12 -> 19200bps @ 4MHz
	ldi	TEMP, 0b00011000;Receiver and Transmitter enabled
	out	UCR, TEMP
	
;Init PORTB
	ser	TEMP
	out	DDRB, TEMP	;PORTB = output
	out	PORTB, TEMP	;All LEDs cleared (LEDs inverted on STK200)

;Init ACK_COUNT
	clr	ACK_COUNT


NO_SYNC:
	clr	CRC_H		;Reset CRC16 
	clr	CRC_L
	
	rcall	RX_LOOP		;Wait for new byte
	cpi	DATA, SYNC
	brne	NO_SYNC		;Received byte was not SYNC
	
	rcall	RX_LOOP		;Wait for new byte
	cpi	DATA, HDB2
	brne	NO_SYNC		;HDB2 not OK for me
	rcall	CRC16
	
	rcall	RX_LOOP		;Wait for new byte
	cpi	DATA, HDB1
	brne	NO_SYNC		;HDB1 not OK for me
	rcall	CRC16
	
	rcall	RX_LOOP		;Wait for new byte
	cpi	DATA, ME
	brne	NO_SYNC		;DAB1 not me
	rcall	CRC16

	rcall	RX_LOOP		;Wait for new byte
	mov	SAB1, DATA	;Save SAB1
	rcall	CRC16
	
	rcall	RX_LOOP		;Wait for new byte
	mov	DB1, DATA	;Save DB1
	rcall	CRC16

	rcall	RX_LOOP		;Wait for new byte
	rcall	CRC16
	rcall	RX_LOOP		;Wait for new byte
	rcall	CRC16
	tst	CRC_H
	brne	SEND_NAK	;CRC16 not 0
	tst	CRC_L
	brne	SEND_NAK	;CRC16 not 0
	
	
; Shows DB1 on LEDs
	com	DB1
	out	PORTB, DB1	;Inverted DB1 to LEDs (LEDs on STK200 are inverted)


; Sends ACK with ACK_COUNT as DB1	
	ldi	DATA, SYNC
	rcall	TX
	
	ldi	DATA, ACK_HDB2
	rcall	CRC16
	rcall	TX
	
	ldi	DATA, ACK_HDB1
	rcall	CRC16
	rcall	TX
	
	mov	DATA, SAB1	;SAB1 of received packet now acts as DAB1 
	rcall	CRC16
	rcall	TX
	
	ldi	DATA, ME	;Nodes adress as SAB1 
	rcall	CRC16
	rcall	TX	
	
	inc	ACK_COUNT
	mov	DATA, ACK_COUNT	;ACK_COUNT as DB1
	rcall	CRC16
	rcall	TX
	
	mov	DATA, CRC_H
	rcall	TX

	mov	DATA, CRC_L
	rcall	TX

	rjmp	NO_SYNC
	

SEND_NAK:
	clr	CRC_H		;Reset CRC16
	clr	CRC_L
	
	ldi	DATA, SYNC
	rcall	TX
	
	ldi	DATA, NAK_HDB2
	rcall	CRC16
	rcall	TX
	
	ldi	DATA, NAK_HDB1
	rcall	CRC16
	rcall	TX
	
	mov	DATA, SAB1	;SAB1 of received packet now acts as DAB1 
	rcall	CRC16
	rcall	TX
	
	ldi	DATA, ME	;Nodes adress as SAB1 
	rcall	CRC16
	rcall	TX

	mov	DATA, CRC_H
	rcall	TX

	mov	DATA, CRC_L
	rcall	TX
	
	rjmp	NO_SYNC
	




;*** RX_LOOP ***
;Returns received byte in DATA
RX_LOOP:
	sbis	USR, RXC
	rjmp	RX_LOOP
	in	DATA, UDR
	ret
	
	
;*** TX ***
;Sends DATA
TX:	sbi	USR, TXC
	out	UDR, DATA
TX_LOOP:
	sbis	USR, TXC
	rjmp	TX_LOOP
	ret





;*** CRC16 ***
;Purpose:
;	Calculate CRC-16 Values. Uses The CCITT-16 Polynomial,
;	Expressed As X^16 + X^12 + X^5 + 1
;Author:
;	John C. Wren
;Origin:
;	http://www.kd4dts.com/download/zip/crc16.zip  [Crc16.asm]
;2000-07-17:
;	Modifyed from MCS-51 to AVR by Gran Nordahl

CRC16:	eor	CRC_H, DATA
	ldi	COUNT, 0x08
CRC16_LOOP:
	lsl	CRC_L
	rol	CRC_H
	brcc	CRC16_SKIP
	ldi	TEMP, 0x10
	eor	CRC_H, TEMP
	ldi	TEMP, 0x21
	eor	CRC_L, TEMP
CRC16_SKIP:
	dec	COUNT
	brne	CRC16_LOOP
	ret
;Vulpes^2